#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_apikey.zmi.interfaces module

This module defines interfaces used by management interface.
"""

from zope.interface import Interface


__docformat__ = 'restructuredtext'


class IAPIKeyConfigurationMenu(Interface):
    """API key configuration menu marker interface"""


class IAPIKeyContainerTable(Interface):
    """API key configuration table marker interface"""



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_apikey.zmi.plugin module

This modules defines main API keys configuration management components.
"""

from zope.interface import Interface

from pyams_auth_apikey.interfaces import IAPIKeyConfiguration
from pyams_auth_apikey.zmi.interfaces import IAPIKeyConfigurationMenu
from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.interfaces.form import IGroup
from pyams_layer.interfaces import IPyAMSLayer
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.base import MANAGE_SECURITY_PERMISSION
from pyams_security_views.zmi import ISecurityMenu
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import get_utility
from pyams_viewlet.manager import viewletmanager_config
from pyams_zmi.form import AdminEditForm, FormGroupChecker
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem


__docformat__ = 'restructuredtext'

from pyams_auth_apikey import _


@viewletmanager_config(name='apikey-configuration.menu',
                       context=ISecurityManager, layer=IAdminLayer,
                       manager=ISecurityMenu, weight=17,
                       provides=IAPIKeyConfigurationMenu,
                       permission=MANAGE_SECURITY_PERMISSION)
class APIKeyConfigurationMenu(NavigationMenuItem):
    """API keys configuration menu"""

    label = _("API keys configuration")
    href = '#apikey-configuration.html'


@ajax_form_config(name='apikey-configuration.html',
                  context=ISecurityManager, layer=IPyAMSLayer,
                  permission=MANAGE_SECURITY_PERMISSION)
class APIKeyConfigurationEditForm(AdminEditForm):
    """API keys configuration edit form"""

    title = _("API keys configuration")

    fields = Fields(Interface)

    def get_content(self):
        """Content getter"""
        return get_utility(ISecurityManager)


@adapter_config(name='apikey-configuration',
                required=(ISecurityManager, IAdminLayer, APIKeyConfigurationEditForm),
                provides=IGroup)
class APIKeyConfigurationGroup(FormGroupChecker):
    """API keys configuration group"""

    fields = Fields(IAPIKeyConfiguration)



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_apikey.zmi module

This module defines management components which are used to handle API keys.
"""

import hashlib
import hmac
import random
import sys
from datetime import datetime, timedelta, timezone

from pyramid.events import subscriber
from zope.interface import Interface, Invalid

from pyams_auth_apikey.interfaces import IAPIKey, IAPIKeyConfiguration
from pyams_auth_apikey.zmi.interfaces import IAPIKeyContainerTable
from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.interfaces import DISPLAY_MODE
from pyams_form.interfaces.form import IAJAXFormRenderer, IDataExtractedEvent, IGroup
from pyams_layer.interfaces import IPyAMSLayer
from pyams_security.interfaces import ISecurityManager, IViewContextPermissionChecker
from pyams_security.interfaces.base import MANAGE_SECURITY_PERMISSION
from pyams_skin.interfaces.view import IModalAddForm, IModalEditForm
from pyams_skin.interfaces.viewlet import IFormHeaderViewletManager
from pyams_skin.viewlet.actions import ContextAddAction
from pyams_skin.viewlet.help import AlertMessage
from pyams_skin.widget.text import TextCopyFieldWidget
from pyams_utils.adapter import ContextAdapter, ContextRequestViewAdapter, adapter_config
from pyams_utils.interfaces import ICacheKeyValue
from pyams_utils.registry import get_utility
from pyams_utils.timezone import tztime
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminModalAddForm, AdminModalEditForm, FormGroupChecker
from pyams_zmi.helper.event import get_json_table_row_add_callback, get_json_table_row_refresh_callback
from pyams_zmi.interfaces import IAdminLayer, TITLE_SPAN_BREAK
from pyams_zmi.interfaces.form import IFormTitle
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.interfaces.viewlet import IToolbarViewletManager
from pyams_zmi.table import TableElementEditor
from pyams_zmi.utils import get_object_label


__docformat__ = 'restructuredtext'

from pyams_auth_apikey import _


@viewlet_config(name='add-apikey.action',
                context=ISecurityManager, layer=IAdminLayer, view=IAPIKeyContainerTable,
                manager=IToolbarViewletManager, weight=10,
                permission=MANAGE_SECURITY_PERMISSION)
class APIKeyAddAction(ContextAddAction):
    """API key add action"""

    label = _("Add API key")
    href = 'add-apikey.html'

    def get_href(self):
        configuration = IAPIKeyConfiguration(self.context)
        return absolute_url(configuration, self.request, self.href)


@ajax_form_config(name='add-apikey.html',
                  context=IAPIKeyConfiguration, layer=IPyAMSLayer,
                  permission=MANAGE_SECURITY_PERMISSION)
class APIKeyAddForm(AdminModalAddForm):
    """API key add form"""

    subtitle = _("New API key")
    legend = _("New API key properties")

    fields = Fields(IAPIKey).omit('hash', 'enabled', 'restrict_referrers', 'allowed_referrers')
    fields['key'].widget_factory = TextCopyFieldWidget

    content_factory = IAPIKey

    def update_widgets(self, prefix=None):
        super().update_widgets(prefix)
        key = self.widgets.get('key')
        if key is not None:
            secret = ''.join((hex(random.randint(0, sys.maxsize))[2:]
                              for i in range(4)))
            seed1 = '-'.join((hex(random.randint(0, sys.maxsize))[2:]
                              for i in range(4)))
            seed2 = '-'.join((hex(random.randint(0, sys.maxsize))[2:]
                              for i in range(4)))
            secret = hmac.new(secret.encode(), seed1.encode(), digestmod=hashlib.sha512)
            secret.update(seed2.encode())
            key.value = secret.hexdigest()
            key.readonly = 'readonly'
        expiration_date = self.widgets.get('expiration_date')
        if expiration_date is not None:
            now = tztime(datetime.now(timezone.utc))
            expiration_date.value = now + timedelta(days=365)

    def update_content(self, obj, data):
        """Update new content properties"""
        changes = super().update_content(obj, data)
        if not obj.name:
            obj.name = ICacheKeyValue(obj)
        return changes

    def add(self, obj):
        """Add API key to container"""
        self.context[obj.name] = obj


@adapter_config(required=(IAPIKeyConfiguration, IAdminLayer, IModalAddForm),
                provides=IFormTitle)
def apikey_add_form_title(context, request, form):
    """API key add form title getter"""
    translate = request.localizer.translate
    manager = get_utility(ISecurityManager)
    return TITLE_SPAN_BREAK.format(
        get_object_label(manager, request, form),
        translate(_("Plug-in: API keys authentication")))


@adapter_config(name='referrers.group',
                required=(IAPIKeyConfiguration, IAdminLayer, APIKeyAddForm),
                provides=IGroup)
class APIKeyAddFormReferrersGroup(FormGroupChecker):
    """API key add form referrers group"""

    fields = Fields(IAPIKey).select('restrict_referrers', 'allowed_referrers')


@viewlet_config(name='add-api-key.header',
                context=IAPIKeyConfiguration, layer=IAdminLayer, view=APIKeyAddForm,
                manager=IFormHeaderViewletManager, weight=1)
class APIKeyAddFormHelp(AlertMessage):
    """API key add form help"""

    _message = _("WARNING: the API key won't be displayed after it's creation!\n"
                 "You must copy it NOW and keep it safe!!")
    message_renderer = 'text'

    status = 'danger'
    icon_class = 'fas fa-radiation'


@subscriber(IDataExtractedEvent, form_selector=APIKeyAddForm)
def handle_apikey_add_form_data(event):
    """Handle new API key data"""
    data = event.data
    name = data.get('name')
    if name and (name in event.form.context):
        event.form.widgets.errors += (Invalid(_("This key name is already used!")),)


@adapter_config(required=(IAPIKeyConfiguration, IAdminLayer, APIKeyAddForm),
                provides=IAJAXFormRenderer)
class APIKeyAddFormRenderer(ContextRequestViewAdapter):
    """API key add form renderer"""

    def render(self, changes):
        """JSON form renderer"""
        if not changes:
            return None
        sm = get_utility(ISecurityManager)
        return {
            'callbacks': [
                get_json_table_row_add_callback(sm, self.request,
                                                IAPIKeyContainerTable, changes)
            ]
        }


@adapter_config(required=(IAPIKey, IAdminLayer, Interface),
                provides=ITableElementEditor)
class APIKeyElementEditor(TableElementEditor):
    """API key table element editor"""


@adapter_config(required=IAPIKey,
                provides=IViewContextPermissionChecker)
class APIKeyPermissionChecker(ContextAdapter):
    """API key permission checker"""

    edit_permission = MANAGE_SECURITY_PERMISSION


@ajax_form_config(name='properties.html',
                  context=IAPIKey, layer=IPyAMSLayer,
                  permission=MANAGE_SECURITY_PERMISSION)
class APIKeyPropertiesEditForm(AdminModalEditForm):
    """API key properties edit form"""

    @property
    def subtitle(self):
        translate = self.request.localizer.translate
        return translate(_("API key: {}")).format(self.context.label)

    legend = _("API key properties")

    fields = Fields(IAPIKey).omit('key', 'enabled', 'restrict_referrers', 'allowed_referrers')

    def update_widgets(self, prefix=None):
        """Widgets update"""
        super().update_widgets(prefix)
        name = self.widgets.get('name')
        if name is not None:
            name.mode = DISPLAY_MODE


@adapter_config(required=(IAPIKey, IAdminLayer, IModalEditForm),
                provides=IFormTitle)
def apikey_edit_form_title(context, request, form):
    """API key add form title getter"""
    translate = request.localizer.translate
    manager = get_utility(ISecurityManager)
    return TITLE_SPAN_BREAK.format(
        get_object_label(manager, request, form),
        translate(_("Plug-in: API keys authentication")))


@adapter_config(name='referrers.group',
                required=(IAPIKey, IAdminLayer, APIKeyPropertiesEditForm),
                provides=IGroup)
class APIKeyPropertiesEditFormReferrersGroup(FormGroupChecker):
    """API key properties edit form referrers group"""

    fields = Fields(IAPIKey).select('restrict_referrers', 'allowed_referrers')


@adapter_config(required=(IAPIKey, IAdminLayer, APIKeyPropertiesEditForm),
                provides=IAJAXFormRenderer)
class APIKeyPropertiesEditFormRenderer(ContextRequestViewAdapter):
    """API key properties edit form renderer"""

    def render(self, changes):
        """JSON form renderer"""
        if not changes:
            return None
        sm = get_utility(ISecurityManager)
        return {
            'status': 'success',
            'message': self.request.localizer.translate(self.view.success_message),
            'callbacks': [
                get_json_table_row_refresh_callback(sm, self.request,
                                                    IAPIKeyContainerTable, self.context)
            ]
        }



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_apikey.interfaces module

"""

from zope.container.constraints import contains
from zope.container.interfaces import IContainer
from zope.interface import Attribute, Interface, Invalid, invariant
from zope.schema import Bool, BytesLine, Datetime, TextLine

from pyams_security.interfaces.plugin import IAuthenticationPlugin, ICredentialsPlugin, IDirectoryPlugin
from pyams_security.schema import PrincipalField
from pyams_utils.schema import TextLineListField

__docformat__ = 'restructuredtext'

from pyams_auth_apikey import _


APIKEY_CONFIGURATION_KEY = 'pyams_auth_apikey.configuration'
"""Main API key configuration key"""

APIKEY_PLUGIN_LABEL = _("API keys authentication plug-in")
"""API keys plugin label"""

APIKEY_PREFIX = 'apikey'
"""API key plugin prefix"""

APIKEY_SALT = 'PyAMS'
"""Default key encoding salt"""


class IAPIKey(Interface):
    """Base API key interface"""

    name = TextLine(title=_("Key name"),
                    description=_("This name must be unique between all API keys; if empty, a random name will be "
                                  "used as name"),
                    required=False)

    key = TextLine(title=_("API key"),
                   description=_("This API key will never be displayed again, just copy it and keep it safe!"),
                   required=False)

    hash = BytesLine(title=_("API key hash"),
                     description=_("This is a hash of the API key and can't be modified"),
                     required=True,
                     readonly=True)

    enabled = Bool(title=_("Enabled key?"),
                   description=_("Select 'no' to disable the key temporarily"),
                   required=True,
                   default=True)

    label = TextLine(title=_("Key label"),
                     description=_("This label will be used to identify the key"),
                     required=False)

    principal_id = PrincipalField(title=_("Associated principal"),
                                  description=_("If defined, this will identify the principal which will be used when "
                                                "a request will be authenticated with this API key"),
                                  required=False)

    @invariant
    def check_principal_id(self):
        """Check principal ID"""
        if self.principal_id and (self.principal_id.startswith(f'{APIKEY_PREFIX}:')):
            raise Invalid(_("Selected principal can't be another API key!"))

    def get_principal(self, request=None):
        """Get principal matching the API key"""

    activation_date = Datetime(title=_("Activation date"),
                               description=_("This key will be enabled only after this date"),
                               required=False)

    expiration_date = Datetime(title=_("Expiration date"),
                               description=_("This key will not be enabled after this date"),
                               required=True)

    restrict_referrers = Bool(title=_("Restrict referrers"),
                              description=_("If this option is enabled, only selected referrers will be enabled"),
                              required=True,
                              default=False)

    allowed_referrers = TextLineListField(title=_("Allowed referrers"),
                                          description=_("Only selected referrers will be allowed to use this API key"),
                                          required=False)

    active = Attribute("Key activity checker")


class IAPIKeyConfiguration(IContainer):
    """API keys plug-in configuration interface"""

    contains(IAPIKey)

    enabled = Bool(title=_("Enabled API key authentication?"),
                   description=_("Enable login via API keys"),
                   required=False,
                   default=False)

    http_header = TextLine(title=_("HTTP header"),
                           description=_("Name of the HTTP header used to provide the API key"),
                           required=True,
                           default='X-API-Key')

    by_hash = Attribute("API keys mapping by key hash")
    by_principal = Attribute("API keys mapping by principal ID")

    def get_apikey(self, request):
        """Extract API key from provided request"""

    def find_active_key(self, request):
        """Get active API key with given ID"""


class IAPIKeyPlugin(ICredentialsPlugin, IAuthenticationPlugin, IDirectoryPlugin):
    """API keys authentication module"""

    configuration = Attribute("API keys configuration")
    enabled = Attribute("Enable API keys authentication?")



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_apikey.plugin module

This is the main security plugin module used for API keys authentication.
"""

from datetime import datetime, timezone
from functools import partial, wraps

from BTrees.OOBTree import OOBTree
from ZODB.POSException import ConnectionStateError
from persistent import Persistent
from zope.container.contained import Contained
from zope.container.folder import Folder
from zope.password.interfaces import IPasswordManager
from zope.schema.fieldproperty import FieldProperty
from zope.traversing.interfaces import ITraversable

from pyams_auth_apikey.interfaces import APIKEY_CONFIGURATION_KEY, APIKEY_PREFIX, APIKEY_SALT, IAPIKey, \
    IAPIKeyConfiguration, IAPIKeyPlugin
from pyams_security.credential import Credentials
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.names import PRINCIPAL_ID_FORMATTER
from pyams_security.interfaces.plugin import IAuthenticationPlugin, ICredentialsPlugin, IDirectoryPlugin
from pyams_security.principal import PrincipalInfo
from pyams_security.utility import get_principal
from pyams_utils.adapter import ContextAdapter, adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.property import ClassPropertyType, classproperty
from pyams_utils.registry import get_pyramid_registry, query_utility, utility_config
from pyams_utils.request import query_request
from pyams_utils.timezone import tztime
from pyams_utils.zodb import volatile_property


__docformat__ = 'restructuredtext'

from pyams_auth_apikey import LOGGER, _


@factory_config(IAPIKey)
class APIKey(Persistent, Contained):
    """API key persistent class"""

    name = FieldProperty(IAPIKey['name'])
    hash = FieldProperty(IAPIKey['hash'])
    _enabled = FieldProperty(IAPIKey['enabled'])
    label = FieldProperty(IAPIKey['label'])
    _principal_id = FieldProperty(IAPIKey['principal_id'])
    restrict_referrers = FieldProperty(IAPIKey['restrict_referrers'])
    allowed_referrers = FieldProperty(IAPIKey['allowed_referrers'])
    _activation_date = FieldProperty(IAPIKey['activation_date'])
    _expiration_date = FieldProperty(IAPIKey['expiration_date'])

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def key(self):
        """API key getter"""
        return None

    @key.setter
    def key(self, value):
        """API key setter, only stores hash"""
        registry = get_pyramid_registry()
        encoder = registry.getUtility(IPasswordManager, name='PBKDF2')
        self.hash = encoder.encodePassword(value, salt=APIKEY_SALT)[8:]

    @property
    def enabled(self):
        """Enabled field getter"""
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        """Enabled field setter"""
        if value != self._enabled:
            self._enabled = value
            del self.active

    @property
    def principal_id(self):
        """Principal ID getter"""
        return self._principal_id

    @principal_id.setter
    def principal_id(self, value):
        """Principal ID setter"""
        plugin = self.__parent__
        if plugin is not None:
            plugin.update_key(self, self._principal_id, value)
        self._principal_id = value

    @property
    def activation_date(self):
        """Activation date field getter"""
        return self._activation_date

    @activation_date.setter
    def activation_date(self, value):
        """Activation date field setter"""
        if value != self._activation_date:
            self._activation_date = value
            del self.active

    @property
    def expiration_date(self):
        """Expiration date field getter"""
        return self._expiration_date

    @expiration_date.setter
    def expiration_date(self, value):
        """Expiration date field setter"""
        if value != self._expiration_date:
            self._expiration_date = value
            del self.active

    @volatile_property
    def active(self):
        """Key activity checker"""
        if not self.enabled:
            return False
        now = tztime(datetime.now(timezone.utc))
        if self.activation_date and (self.activation_date > now):
            return False
        if self.expiration_date and (self.expiration_date < now):
            return False
        return True

    def get_principal(self, request=None, allow_redirect=True):
        """Get principal matching this API key"""
        if not self.active:
            return None
        if request is None:
            request = query_request()
        if self.restrict_referrers:
            origin = request.headers.get('Origin', request.host_url)
            if not ((origin == request.host_url) or (origin in self.allowed_referrers or ())):
                return None
        if self.principal_id and allow_redirect:
            return get_principal(request, self.principal_id)
        translate = request.localizer.translate
        return PrincipalInfo(id=PRINCIPAL_ID_FORMATTER.format(prefix=APIKEY_PREFIX, login=self.name),
                             title=translate(_("API key: {}")).format(self.label))


def check_enabled(func=None, *, default=None):
    """Decorator to check for enabled plugin"""
    if func is None:
        return partial(check_enabled, default=default)

    @wraps(func)
    def wrapper(plugin, *args, **kwargs):
        if not plugin.enabled:
            return default() if default is not None else None
        return func(plugin, *args, **kwargs)
    return wrapper


def check_prefix(func=None, *, default=None):
    """Decorator to check for principal ID prefix"""
    if func is None:
        return partial(check_prefix, default=default)

    @wraps(func)
    def wrapper(plugin, principal_id, *args, **kwargs):
        if not principal_id.startswith(f'{APIKEY_PREFIX}:'):
            return default() if default is not None else None
        return func(plugin, principal_id, *args, **kwargs)
    return wrapper


@factory_config(IAPIKeyConfiguration)
class APIKeyConfiguration(Folder):
    """API keys container folder"""

    enabled = FieldProperty(IAPIKeyConfiguration['enabled'])
    http_header = FieldProperty(IAPIKeyConfiguration['http_header'])

    by_hash = None
    by_principal = None

    def _newContainerData(self):
        """Create new container data"""
        data = super()._newContainerData()
        self.by_hash = OOBTree()
        self.by_principal = OOBTree()
        return data

    def __setitem__(self, key, value):
        """Store new API key"""
        super().__setitem__(key, value)
        self.by_hash[value.hash] = value
        if value.principal_id:
            self.by_principal.setdefault(value.principal_id, []).append(value)

    def __delitem__(self, key):
        """Del API key"""
        apikey = self.get(key)
        if apikey is not None:
            del self.by_hash[apikey.hash]
            if apikey.principal_id:
                keys = self.by_principal.get(apikey.principal_id)
                if apikey in keys:
                    keys.remove(apikey)
                if keys:
                    self.by_principal[apikey.principal_id] = keys
                else:
                    del self.by_principal[apikey.principal_id]
        super().__delitem__(key)

    def update_key(self, apikey, old_principal_id, new_principal_id):
        """Update key"""
        keys = self.by_principal.get(old_principal_id)
        if keys:
            if apikey in keys:
                keys.remove(apikey)
            if keys:
                self.by_principal[old_principal_id] = keys
            else:
                del self.by_principal[old_principal_id]
        if new_principal_id:
            self.by_principal.setdefault(new_principal_id, []).append(apikey)

    @check_enabled
    def get_apikey_header(self, request):
        """Extract API key from request header"""
        return request.headers.get(self.http_header)

    @staticmethod
    def get_key_hash(key):
        """Convert key to it's hash"""
        registry = get_pyramid_registry()
        encoder = registry.getUtility(IPasswordManager, name='PBKDF2')
        return encoder.encodePassword(key, salt=APIKEY_SALT)[8:]

    @check_enabled
    def find_active_key(self, key):
        """Find API key matching provided key"""
        hash = self.get_key_hash(key)
        apikey = self.by_hash.get(hash)
        if (apikey is None) or (not apikey.active):
            return None
        LOGGER.debug("  > Found API key: %s", apikey.name)
        return apikey

    def extract_credentials(self, request):
        """Extract credentials from request"""
        header = self.get_apikey_header(request)
        if header is None:
            return None
        apikey = self.find_active_key(header)
        if apikey is None:
            return None
        return Credentials(APIKEY_PREFIX, f'{APIKEY_PREFIX}:{apikey.name}')

    @check_enabled
    def authenticate(self, credentials, request):
        """Authenticate provided credentials against active API keys"""
        prefix, name = credentials.id.split(':', 1)
        if prefix != APIKEY_PREFIX:
            return None
        apikey = self.get(name)
        if (apikey is None) or (not apikey.active):
            return None
        principal = apikey.get_principal(request)
        if principal is not None:
            return principal.id
        return None

    @check_enabled
    @check_prefix
    def get_principal(self, principal_id, info=True):
        """Returns real principal matching given ID, or None"""
        prefix, name = principal_id.split(':', 1)
        apikey = self.get(name)
        if apikey is None:
            return None
        if info:
            return apikey.get_principal()
        return apikey

    @check_enabled(default=set)
    @check_prefix(default=set)
    def get_all_principals(self, principal_id):
        """Returns all principals matching given principal ID"""
        result = set()
        prefix, name = principal_id.split(':', 1)
        apikey = self.get(name)
        if (apikey is not None) and apikey.active:
            result.add(principal_id)
            principal = apikey.get_principal()
            if principal is not None:
                result.add(principal.id)
        return result

    @check_enabled
    def find_principals(self, query, exact_match=False):
        """Find principals matching given query"""
        if not query:
            return
        query = query.lower()
        for apikey in self.values():
            if not apikey.active:
                continue
            for attr in (apikey.name, apikey.label):
                if not attr:
                    continue
                if (exact_match and query == attr.lower()) or \
                        (not exact_match and query in attr.lower()):
                    yield apikey.get_principal(allow_redirect=False)
                    break


@adapter_config(required=ISecurityManager,
                provides=IAPIKeyConfiguration)
def security_manager_apikey_configuration(context):
    """Security manager API keys configuration"""
    return get_annotation_adapter(context, APIKEY_CONFIGURATION_KEY, IAPIKeyConfiguration,
                                  name='++apikey-config++')


@adapter_config(name='apikey-config',
                required=ISecurityManager,
                provides=ITraversable)
class SecurityManagerAPIKeyConfigurationTraverser(ContextAdapter):
    """Security manager API key configuration traverser"""

    def traverse(self, name, furtherpath=None):  # pylint: disable=unused-argument
        """Traverse to API key configuration"""
        return IAPIKeyConfiguration(self.context)


@utility_config(provides=IAPIKeyPlugin)
@utility_config(name='apikey', provides=ICredentialsPlugin)
@utility_config(name='apikey', provides=IAuthenticationPlugin)
@utility_config(name='apikey', provides=IDirectoryPlugin)
class APIKeyPlugin(metaclass=ClassPropertyType):
    """API keys plug-in"""

    prefix = APIKEY_PREFIX
    title = _("API keys authentication")

    @classproperty
    def configuration(cls):  # pylint: disable=no-self-argument,no-self-use
        """API keys configuration getter"""
        try:
            sm = query_utility(ISecurityManager)  # pylint: disable=invalid-name
            if sm is not None:
                return IAPIKeyConfiguration(sm)
        except ConnectionStateError:
            return None
        return None

    @classproperty
    def enabled(cls):  # pylint: disable=no-self-argument,no-self-use
        """Check if API keys authentication is enabled"""
        configuration = cls.configuration
        try:
            return configuration.enabled if (configuration is not None) else False
        except ConnectionStateError:
            return False

    def extract_credentials(self, request, **kwargs):  # pylint: disable=unused-argument
        """Extract credentials from HTTP header"""
        configuration = self.configuration
        if configuration is None:
            return None
        return configuration.extract_credentials(request)

    def authenticate(self, credentials, request):
        """Authenticate provided credentials"""
        if credentials.prefix != self.prefix:
            return None
        configuration = self.configuration
        if configuration is None:
            return None
        return configuration.authenticate(credentials, request)

    def get_principal(self, principal_id, info=True):
        """Returns real principal matching given ID, or None"""
        configuration = self.configuration
        if configuration is None:
            return None
        return configuration.get_principal(principal_id, info)

    def get_all_principals(self, principal_id):
        """Returns all principals matching given principal ID"""
        configuration = self.configuration
        if configuration is None:
            return set()
        return configuration.get_all_principals(principal_id) or set()

    def find_principals(self, query, exact_match=False):
        """Find principals matching given query"""
        configuration = self.configuration
        if configuration is None:
            return
        yield from configuration.find_principals(query, exact_match) or set()



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_apikey.zmi.container module

This module defines components which are used to manage API keys containers.
"""

from pyramid.decorator import reify
from pyramid.view import view_config

from pyams_auth_apikey.interfaces import IAPIKeyConfiguration
from pyams_auth_apikey.zmi.interfaces import IAPIKeyContainerTable, IAPIKeyConfigurationMenu
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.base import MANAGE_SECURITY_PERMISSION
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.date import format_datetime
from pyams_utils.factory import factory_config
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.helper.container import delete_container_element, switch_element_attribute
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.table import AttributeSwitcherColumn, I18nColumnMixin, Table, TableAdminView, TrashColumn
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem


__docformat__ = 'restructuredtext'

from pyams_auth_apikey import _


@viewlet_config(name='apikey-container.menu',
                context=ISecurityManager, layer=IAdminLayer,
                manager=IAPIKeyConfigurationMenu, weight=10,
                permission=MANAGE_SECURITY_PERMISSION)
class APIKeyContainerMenu(NavigationMenuItem):
    """API keys container menu"""

    label = _("API keys")
    href = '#apikey-container.html'


@factory_config(IAPIKeyContainerTable)
class APIKeyContainerTable(Table):
    """API key container table"""

    display_if_empty = True

    @reify
    def data_attributes(self):
        attributes = super().data_attributes
        configuration = IAPIKeyConfiguration(self.context)
        attributes['table'].update({
            'data-ams-order': '1,asc',
            'data-ams-location': absolute_url(configuration, self.request)
        })
        return attributes


@adapter_config(required=(ISecurityManager, IAdminLayer, APIKeyContainerTable),
                provides=IValues)
class APIKeyContainerTableValues(ContextRequestViewAdapter):
    """API keys container table values adapter"""

    @property
    def values(self):
        """Table values getter"""
        yield from IAPIKeyConfiguration(self.context).values()


@adapter_config(name='enabled',
                required=(ISecurityManager, IAdminLayer, APIKeyContainerTable),
                provides=IColumn)
class APIKeyEnabledSwitcherColumn(AttributeSwitcherColumn):
    """API key enabled column"""

    hint = _("Click icon to enable or disable this API key")

    attribute_name = 'enabled'
    attribute_switcher = 'switch-enabled-key.json'

    icon_on_class = 'fas fa-shop-lock'
    icon_off_class = 'fas fa-shop-slash text-danger'

    weight = 10


@view_config(name='switch-enabled-key.json',
             context=IAPIKeyConfiguration, request_type=IPyAMSLayer,
             renderer='json', xhr=True)
def switch_enabled_key(request):
    """Switch enabled API key"""
    return switch_element_attribute(request)


@adapter_config(name='name',
                required=(ISecurityManager, IAdminLayer, APIKeyContainerTable),
                provides=IColumn)
class APIKeyNameColumn(I18nColumnMixin, GetAttrColumn):
    """API key container name column"""

    i18n_header = _("Name")
    attr_name = 'name'

    weight = 20


@adapter_config(name='label',
                required=(ISecurityManager, IAdminLayer, APIKeyContainerTable),
                provides=IColumn)
class APIKeyLabelColumn(I18nColumnMixin, GetAttrColumn):
    """API key container label column"""

    i18n_header = _("Label")
    attr_name = 'label'

    weight = 30

    def get_value(self, obj):
        """Value getter"""
        return super().get_value(obj) or '--'


@adapter_config(name='principal',
                required=(ISecurityManager, IAdminLayer, APIKeyContainerTable),
                provides=IColumn)
class APIKeyPrincipalColumn(I18nColumnMixin, GetAttrColumn):
    """API key container principal column"""

    i18n_header = _("Principal")

    weight = 40

    def get_value(self, obj):
        """Value getter"""
        principal = obj.get_principal(self.request)
        return principal.title if principal is not None else '--'


@adapter_config(name='activation_date',
                required=(ISecurityManager, IAdminLayer, APIKeyContainerTable),
                provides=IColumn)
class APIKeyActivationDateColumn(I18nColumnMixin, GetAttrColumn):
    """API key container activation date column"""

    i18n_header = _("Activation date")
    attr_name = 'activation_date'

    weight = 50

    def get_value(self, obj):
        """Activation date getter"""
        activation_date = super().get_value(obj)
        return format_datetime(activation_date)


@adapter_config(name='expiration_date',
                required=(ISecurityManager, IAdminLayer, APIKeyContainerTable),
                provides=IColumn)
class APIKeyExpirationDateColumn(I18nColumnMixin, GetAttrColumn):
    """API key container expiration date column"""

    i18n_header = _("Expiration date")
    attr_name = 'expiration_date'

    weight = 60

    def get_value(self, obj):
        """Expiration date getter"""
        expiration_date = super().get_value(obj)
        return format_datetime(expiration_date)


@adapter_config(name='trash',
                required=(ISecurityManager, IAdminLayer, APIKeyContainerTable),
                provides=IColumn)
class APIKeyTrashColumn(TrashColumn):
    """API key container trash column"""


@view_config(name='delete-element.json',
             context=IAPIKeyConfiguration, request_type=IPyAMSLayer,
             permission=MANAGE_SECURITY_PERMISSION, renderer='json', xhr=True)
def delete_apikey(request):
    """API key delete view"""
    return delete_container_element(request)


@pagelet_config(name='apikey-container.html',
                context=ISecurityManager, layer=IPyAMSLayer,
                permission=MANAGE_SECURITY_PERMISSION, xhr=True)
class APIKeyContainerView(TableAdminView):
    """API key container view"""

    table_label = _("API keys")
    table_class = IAPIKeyContainerTable
